# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent run templates."""

import copy
import itertools
import re
from abc import ABC
from datetime import datetime
from typing import Any, List, Optional, Set, Tuple, Union, cast
from uuid import uuid4

import marshmallow
from werkzeug.utils import secure_filename

from renku.core import errors
from renku.core.util.datetime8601 import local_now
from renku.domain_model.provenance.agent import Person
from renku.domain_model.provenance.annotation import Annotation
from renku.domain_model.workflow.parameter import (
    CommandInput,
    CommandOutput,
    CommandParameter,
    CommandParameterBase,
    HiddenInput,
)
from renku.infrastructure.database import Persistent

MAX_GENERATED_NAME_LENGTH = 25


class AbstractPlan(Persistent, ABC):
    """Abstract base class for all plans."""

    date_created: datetime
    creators: List[Person] = list()

    def __init__(
        self,
        *,
        description: Optional[str] = None,
        id: str,
        date_created: Optional[datetime] = None,
        date_modified: Optional[datetime] = None,
        date_removed: Optional[datetime] = None,
        keywords: Optional[List[str]] = None,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        derived_from: Optional[str] = None,
        creators: Optional[List[Person]] = None,
    ):
        self.description: Optional[str] = description
        self.id: str = id
        self.date_created: datetime = date_created or local_now()
        self.date_modified: datetime = date_modified or local_now()
        self.date_removed: Optional[datetime] = date_removed
        self.keywords: List[str] = keywords or []

        if creators:
            self.creators = creators

        self.project_id: Optional[str] = project_id
        self.derived_from: Optional[str] = derived_from

        if name is None:
            self.name: str = self._get_default_name()
        else:
            self.validate_name(name)
            self.name = name

    def __repr__(self):
        return f"<{self.__class__.__name__} '{self.name}'>"

    @property
    def deleted(self) -> bool:
        """True if plan is deleted."""
        return self.date_removed is not None

    @staticmethod
    def generate_id(*, uuid: Optional[str] = None, **_) -> str:
        """Generate an identifier for Plan."""
        uuid = uuid or uuid4().hex
        return f"/plans/{uuid}"

    def _get_default_name(self) -> str:
        name = "-".join(str(a).replace(".", "_") for a in self.to_argv())
        if not name:
            return uuid4().hex[:MAX_GENERATED_NAME_LENGTH]

        name = secure_filename(name)
        rand_length = 5
        return f"{name[:MAX_GENERATED_NAME_LENGTH - rand_length -1]}-{uuid4().hex[:rand_length]}"

    @staticmethod
    def validate_name(name: str):
        """Check a name for invalid characters."""
        validate_plan_name(name=name)

    def assign_new_id(self, *, uuid: Optional[str] = None) -> str:
        """Assign a new UUID.

        This is required only when there is another plan which is exactly the same except the parameters' list.
        """
        current_uuid = self._extract_uuid()
        new_uuid = uuid or uuid4().hex
        self.id = self.id.replace(current_uuid, new_uuid)

        # NOTE: We also need to re-assign the _p_oid since identifier has changed
        self.reassign_oid()

        return new_uuid

    def _extract_uuid(self) -> str:
        path_start = self.id.find("/plans/")
        return self.id[path_start + len("/plans/") :]

    def resolve_mapping_path(self, mapping_path: str) -> Tuple[CommandParameterBase, "Plan"]:
        """Resolve a mapping path to its reference parameter."""
        raise NotImplementedError()

    def resolve_direct_reference(self, reference: str) -> CommandParameterBase:
        """Resolve a direct parameter reference."""
        raise NotImplementedError()

    def find_parameter(self, parameter: CommandParameterBase) -> bool:
        """Find if a parameter exists on this plan."""
        raise NotImplementedError()

    def find_parameter_workflow(self, parameter: CommandParameterBase) -> Optional["AbstractPlan"]:
        """Return the workflow a parameter belongs to."""
        raise NotImplementedError()

    def derive(self, creator: Optional[Person] = None) -> "AbstractPlan":
        """Create a new ``AbstractPlan`` that is derived from self."""
        raise NotImplementedError()

    def is_derivation(self) -> bool:
        """Return if an ``AbstractPlan`` has correct derived_from."""
        raise NotImplementedError()

    def delete(self, when: Optional[datetime] = None):
        """Mark a plan as deleted.

        NOTE: Don't call this function for deleting plans since it doesn't delete the whole plan derivatives chain. Use
        renku.core.workflow.plan::remove_plan instead.
        """
        when = when or local_now()

        self.unfreeze()
        self.date_removed = when or local_now()
        self.freeze()


class Plan(AbstractPlan):
    """Represent a `renku run` execution template."""

    annotations: List[Annotation] = list()

    hidden_inputs: List[HiddenInput] = list()
    """Includes a list of dependencies that are defined by Renku and should be hidden from users."""

    def __init__(
        self,
        *,
        annotations: Optional[List[Annotation]] = None,
        command: str,
        creators: Optional[List[Person]] = None,
        date_created: Optional[datetime] = None,
        date_modified: Optional[datetime] = None,
        derived_from: Optional[str] = None,
        description: Optional[str] = None,
        hidden_inputs: Optional[List[HiddenInput]] = None,
        id: str,
        inputs: Optional[List[CommandInput]] = None,
        date_removed: Optional[datetime] = None,
        keywords: Optional[List[str]] = None,
        name: Optional[str] = None,
        outputs: Optional[List[CommandOutput]] = None,
        parameters: Optional[List[CommandParameter]] = None,
        project_id: Optional[str] = None,
        success_codes: Optional[List[int]] = None,
    ):
        self.annotations: List[Annotation] = annotations or []
        self.command: str = command
        self.hidden_inputs: List[HiddenInput] = hidden_inputs or []
        self.inputs: List[CommandInput] = inputs or []
        self.outputs: List[CommandOutput] = outputs or []
        self.parameters: List[CommandParameter] = parameters or []
        self.success_codes: List[int] = success_codes or []
        super().__init__(
            id=id,
            description=description,
            date_created=date_created,
            date_modified=date_modified,
            date_removed=date_removed,
            keywords=keywords,
            name=name,
            project_id=project_id,
            derived_from=derived_from,
            creators=creators,
        )

        # NOTE: Validate plan
        duplicates = get_duplicate_arguments_names(plan=self)
        if duplicates:
            duplicates_string = ", ".join(duplicates)
            raise errors.ParameterError(f"Duplicate input, output or parameter names found: {duplicates_string}")

    @property
    def keywords_csv(self) -> str:
        """Comma-separated list of keywords associated with workflow."""
        return ", ".join(self.keywords)

    def is_equal_to(self, other: "Plan") -> bool:
        """Return true if plan hasn't changed from the other plan."""

        def are_equal_with_order(values, other_values):
            return len(values) == len(other_values) and all(s.is_equal_to(o) for s, o in zip(values, other_values))

        def are_equal(values, other_values):
            return len(values) == len(other_values) and set(values) == set(other_values)

        # TODO: Include ``annotations`` if it is added to the workflow definition file
        return (
            self.name == other.name
            and self.description == other.description
            and self.project_id == other.project_id
            and self.command == other.command
            and are_equal(self.success_codes, other.success_codes)
            and are_equal(self.keywords, other.keywords)
            and are_equal(self.creators, other.creators)
            and are_equal_with_order(self.inputs, other.inputs)
            and are_equal_with_order(self.outputs, other.outputs)
            and are_equal_with_order(self.parameters, other.parameters)
        )

    def resolve_mapping_path(self, mapping_path: str) -> Tuple[CommandParameterBase, "Plan"]:
        """Resolve a mapping path to its reference parameter."""
        parts = mapping_path.split(".", maxsplit=1)

        if len(parts) > 1:
            raise errors.ParameterNotFoundError(mapping_path, self.name)

        return self.resolve_direct_reference(parts[0]), self

    def resolve_direct_reference(self, reference: str) -> CommandParameterBase:
        """Resolve a direct parameter reference."""
        try:
            if reference.startswith("@input"):
                return self.inputs[int(reference[6:]) - 1]
            elif reference.startswith("@output"):
                return self.outputs[int(reference[7:]) - 1]
            elif reference.startswith("@param"):
                return self.parameters[int(reference[6:]) - 1]
        except (ValueError, IndexError):
            raise errors.ParameterNotFoundError(reference, self.name)

        for parameter in self.inputs + self.outputs + self.parameters:  # type: ignore
            if parameter.name == reference:
                return parameter

        raise errors.ParameterNotFoundError(reference, self.name)

    def find_parameter(self, parameter: CommandParameterBase) -> bool:
        """Find if a parameter exists on this plan."""
        return any(
            parameter.id == p.id and parameter.actual_value == p.actual_value
            for p in self.inputs + self.outputs + self.parameters  # type: ignore
        )

    def get_parameter_path(self, parameter: CommandParameterBase):
        """Get the path to a parameter inside this plan."""
        if self.find_parameter(parameter):
            return [self]

        return None

    def get_parameter_by_id(self, parameter_id: str) -> Optional[CommandParameterBase]:
        """Get a parameter on this plan by id."""
        return next(
            (p for p in self.inputs + self.outputs + self.parameters if parameter_id == p.id), None  # type: ignore
        )

    def find_parameter_workflow(self, parameter: CommandParameterBase) -> Optional["Plan"]:
        """Return the workflow a parameter belongs to."""
        if self.find_parameter(parameter):
            return self
        return None

    def assign_new_id(self, *, uuid: Optional[str] = None):
        """Assign a new UUID.

        This is required only when there is another plan which is exactly the same except the parameters' list.
        """
        new_uuid = super().assign_new_id(uuid=uuid)
        current_uuid = self._extract_uuid()
        if hasattr(self, "parameters"):
            self.parameters = copy.deepcopy(self.parameters)

            for a in self.parameters:
                a.id = a.id.replace(current_uuid, new_uuid)

    def derive(self, creator: Optional[Person] = None) -> "Plan":
        """Create a new ``Plan`` that is derived from self."""
        derived = copy.copy(self)
        derived.derived_from = self.id
        derived.date_modified = local_now()
        derived.parameters = self.parameters.copy()
        derived.inputs = self.inputs.copy()
        derived.keywords = copy.deepcopy(self.keywords)
        derived.outputs = self.outputs.copy()
        derived.success_codes = self.success_codes.copy()
        derived.creators = self.creators.copy()
        derived.assign_new_id()

        if creator and hasattr(creator, "email") and not any(c for c in self.creators if c.email == creator.email):
            self.creators.append(creator)

        return derived

    def is_derivation(self) -> bool:
        """Return if an ``Plan`` has correct derived_from."""
        return self.derived_from is not None and self.id != self.derived_from

    def to_argv(self, with_streams: bool = False, quote_string: bool = True) -> List[Any]:
        """Convert a Plan into argv list."""
        arguments = itertools.chain(self.inputs, self.outputs, self.parameters)
        filtered_arguments = filter(lambda a: a.position is not None and not getattr(a, "mapped_to", None), arguments)
        arguments = sorted(filtered_arguments, key=lambda a: a.position)  # type: ignore

        argv = self.command.split(" ") if self.command else []
        argv.extend(e for a in arguments for e in a.to_argv(quote_string=quote_string))

        if with_streams:
            arguments = itertools.chain(self.inputs, self.outputs, self.parameters)
            filtered_arguments = filter(lambda a: a.position is not None and getattr(a, "mapped_to", None), arguments)
            argv.extend(a.to_stream_representation() for a in filtered_arguments)  # type: ignore

        return argv

    def set_parameters_from_strings(self, params_strings: List[str]) -> None:
        """Set parameters by parsing parameters strings."""
        for param_string in params_strings:
            name, value = param_string.split("=", maxsplit=1)
            found = False
            for collection in cast(List[List[CommandParameterBase]], [self.inputs, self.outputs, self.parameters]):
                for i, param in enumerate(collection):
                    if param.name == name:
                        new_param = param.derive(plan_id=self.id)
                        new_param.default_value = value
                        collection[i] = new_param
                        found = True
                        break
                if found:
                    break
            else:
                self.parameters.append(
                    CommandParameter(default_value=value, id=CommandParameter.generate_id(plan_id=self.id), name=name)
                )

    def copy(self):
        """Create a copy of this plan.

        Required where a plan is used several times in a workflow, but we need to set different values on them.
        """
        return copy.deepcopy(self)

    def get_field_by_id(self, id: str) -> Union[CommandInput, CommandOutput, CommandParameter]:
        """Return an in Input/Output/Parameter by its id."""
        for field in itertools.chain(self.inputs, self.outputs, self.parameters):
            if field.id == id:
                return field  # type: ignore

        raise errors.ParameterError(f"Parameter {id} not found on plan {self.id}.")


class PlanDetailsJson(marshmallow.Schema):
    """Serialize a plan to a response object."""

    name = marshmallow.fields.String(required=True)
    full_command = marshmallow.fields.String(data_key="command")
    derived_from = marshmallow.fields.String()
    description = marshmallow.fields.String()
    keywords = marshmallow.fields.List(marshmallow.fields.String())
    id = marshmallow.fields.String()


def get_duplicate_arguments_names(plan: Plan) -> List[str]:
    """Return a list of duplicate inputs/outputs/parameters names in a plan."""
    all_names = [p.name for p in itertools.chain(plan.inputs, plan.outputs, plan.parameters) if p.name]
    seen: Set[str] = set()

    duplicates: List[str] = []
    for n in all_names:
        if n in seen:
            duplicates.append(n)
        else:
            seen.add(n)

    return duplicates


def validate_plan_name(name: str, extra_valid_characters: str = "_-"):
    """Check a name for invalid characters."""
    if not re.match(f"^[a-zA-Z0-9][a-zA-Z0-9{extra_valid_characters}]+$", name):
        raise errors.ParameterError(
            f"Name '{name}' contains illegal characters. Only English letters, numbers, _ and - are allowed."
        )



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Entities JSON-LD schemes."""

from renku.command.schema.calamus import JsonLDSchema, Nested, fields, prov, renku
from renku.domain_model.entity import Collection, Entity


class EntitySchema(JsonLDSchema):
    """Entity Schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Entity]
        model = Entity

    checksum = fields.String(renku.checksum, load_default=None)
    id = fields.Id()
    path = fields.String(prov.atLocation)


class CollectionSchema(EntitySchema):
    """Entity Schema."""

    class Meta:
        """Meta class."""

        rdf_type = prov.Collection
        model = Collection

    members = Nested(prov.hadMember, [EntitySchema, "CollectionSchema"], many=True)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interactive session business logic."""

import os
import shutil
import textwrap
from pathlib import Path
from typing import List, NamedTuple, Optional

from pydantic import validate_arguments

from renku.core import errors
from renku.core.config import get_value
from renku.core.plugin.session import get_supported_hibernating_session_providers, get_supported_session_providers
from renku.core.session.utils import get_image_repository_host, get_renku_project_name
from renku.core.util import communication
from renku.core.util.os import safe_read_yaml
from renku.core.util.ssh import SystemSSHConfig, generate_ssh_keys
from renku.domain_model.session import IHibernatingSessionProvider, ISessionProvider, Session, SessionStopStatus


def _safe_get_provider(provider: str) -> ISessionProvider:
    try:
        return next(p for p in get_supported_session_providers() if p.name == provider)
    except StopIteration:
        raise errors.ParameterError(f"Session provider '{provider}' is not available!")


class SessionList(NamedTuple):
    """Session list return."""

    sessions: List[Session]
    all_local: bool
    warning_messages: List[str]


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def search_sessions(name: str, provider: Optional[str] = None) -> List[str]:
    """Get all sessions that their name starts with the given name.

    Args:
        name(str): The name to search for.
        provider(Optional[str]): Name of the session provider to use (Default value = None).

    Returns:
        All sessions whose name starts with ``name``.
    """
    sessions = session_list(provider=provider).sessions
    name = name.lower()
    return [s.id for s in sessions if s.id.lower().startswith(name)]


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def search_session_providers(name: str) -> List[str]:
    """Get all session providers that their name starts with the given name.

    Args:
        name(str): The name to search for.

    Returns:
        All session providers whose name starts with ``name``.
    """
    from renku.core.plugin.session import get_supported_session_providers

    name = name.lower()
    return [p.name for p in get_supported_session_providers() if p.name.lower().startswith(name)]


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def search_hibernating_session_providers(name: str) -> List[str]:
    """Get all session providers that support hibernation and their name starts with the given name.

    Args:
        name(str): The name to search for.

    Returns:
        All session providers whose name starts with ``name``.
    """
    from renku.core.plugin.session import get_supported_hibernating_session_providers

    name = name.lower()
    return [p.name for p in get_supported_hibernating_session_providers() if p.name.lower().startswith(name)]


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def session_list(*, provider: Optional[str] = None) -> SessionList:
    """List interactive sessions.

    Args:
        provider(Optional[str]): Name of the session provider to use (Default value = None).

    Returns:
        The list of sessions, whether they're all local sessions and potential warnings raised.
    """

    def list_sessions(session_provider: ISessionProvider) -> List[Session]:
        try:
            return session_provider.session_list(project_name=project_name)
        except errors.RenkulabSessionGetUrlError:
            if provider:
                raise
            return []

    project_name = get_renku_project_name()
    providers = [_safe_get_provider(provider)] if provider else get_supported_session_providers()

    all_sessions = []
    warning_messages = []
    all_local = True
    for session_provider in sorted(providers, key=lambda p: p.priority):
        try:
            sessions = list_sessions(session_provider)
        except errors.RenkuException as e:
            warning_messages.append(f"Cannot get sessions list from '{session_provider.name}': {e}")
        else:
            if session_provider.is_remote_provider():
                all_local = False
            all_sessions.extend(sessions)

    return SessionList(all_sessions, all_local, warning_messages)


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def session_start(
    config_path: Optional[str],
    provider: str,
    image_name: Optional[str] = None,
    cpu_request: Optional[float] = None,
    mem_request: Optional[str] = None,
    disk_request: Optional[str] = None,
    gpu_request: Optional[str] = None,
    **kwargs,
):
    """Start interactive session.

    Args:
        config_path(str, optional): Path to config YAML.
        provider(str, optional): Name of the session provider to use.
        image_name(str, optional): Image to start.
        cpu_request(float, optional): Number of CPUs to request.
        mem_request(str, optional): Size of memory to request.
        disk_request(str, optional): Size of disk to request (if supported by provider).
        gpu_request(str, optional): Number of GPUs to request.
    """
    from renku.domain_model.project_context import project_context

    if project_context.repository.head.detached:
        raise errors.SessionStartError("Cannot start a session from a detached HEAD. Check out a branch first.")

    # NOTE: The Docker client in Python requires the parameters below to be a list and will fail with a tuple.
    # Click will convert parameters with the flag "many" set to True to tuples.
    kwargs["security_opt"] = list(kwargs.get("security_opt", []))
    kwargs["device_cgroup_rules"] = list(kwargs.get("device_cgroup_rules", []))

    pinned_image = get_value("interactive", "image")
    if pinned_image and image_name is None:
        image_name = pinned_image

    provider_api = _safe_get_provider(provider)
    config = safe_read_yaml(config_path) if config_path else dict()

    provider_api.pre_start_checks(**kwargs)

    project_name = get_renku_project_name()
    if image_name is None:
        tag = project_context.repository.head.commit.hexsha[:7]
        repo_host = get_image_repository_host()
        image_name = f"{project_name}:{tag}"
        if repo_host:
            image_name = f"{repo_host}/{image_name}"

    force_build_image = provider_api.force_build_image(**kwargs)

    if not force_build_image and not provider_api.find_image(image_name, config):
        communication.confirm(
            f"The container image '{image_name}' does not exist. Would you like to build it using {provider}?",
            abort=True,
        )
        force_build_image = True

    if force_build_image:
        with communication.busy(msg=f"Building image {image_name}"):
            provider_api.build_image(project_context.dockerfile_path.parent, image_name, config)
        communication.echo(f"Image {image_name} built successfully.")

    # set resource settings
    cpu_limit = cpu_request or get_value("interactive", "cpu_request")

    if cpu_limit is not None:
        try:
            cpu_limit = float(cpu_limit)
        except ValueError:
            raise errors.SessionStartError(f"Invalid value for cpu_request (must be float): {cpu_limit}")

    disk_limit = disk_request or get_value("interactive", "disk_request")
    mem_limit = mem_request or get_value("interactive", "mem_request")
    gpu = gpu_request or get_value("interactive", "gpu_request")

    with communication.busy(msg="Waiting for session to start..."):
        provider_message, warning_message = provider_api.session_start(
            config=config,
            project_name=project_name,
            image_name=image_name,
            cpu_request=cpu_limit,
            mem_request=mem_limit,
            disk_request=disk_limit,
            gpu_request=gpu,
            **kwargs,
        )

    if warning_message:
        communication.warn(warning_message)
    communication.echo(provider_message)


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def session_stop(session_name: Optional[str], stop_all: bool = False, provider: Optional[str] = None):
    """Stop interactive session.

    Args:
        session_name(Optional[str]): Name of the session to open.
        stop_all(bool): Whether to stop all sessions or just the specified one.
        provider(Optional[str]): Name of the session provider to use.
    """

    def stop_sessions(session_provider: ISessionProvider) -> SessionStopStatus:
        try:
            return session_provider.session_stop(
                project_name=project_name, session_name=session_name, stop_all=stop_all
            )
        except errors.RenkulabSessionGetUrlError as e:
            if provider:
                raise
            communication.warn(f"Didn't stop any renkulab sessions: {e}")
            return SessionStopStatus.SUCCESSFUL
        except errors.DockerError as e:
            if provider:
                raise
            communication.warn(f"Didn't stop any docker sessions: {e}")
            return SessionStopStatus.SUCCESSFUL

    session_detail = "all sessions" if stop_all else f"session {session_name}" if session_name else "session"
    project_name = get_renku_project_name()

    providers = [_safe_get_provider(provider)] if provider else get_supported_session_providers()

    statues = []
    warning_messages = []
    with communication.busy(msg=f"Waiting for {session_detail} to stop..."):
        for session_provider in sorted(providers, key=lambda p: p.priority):
            try:
                status = stop_sessions(session_provider)
            except errors.RenkuException as e:
                warning_messages.append(f"Cannot stop sessions in provider '{session_provider.name}': {e}")
            else:
                statues.append(status)

            # NOTE: The given session name was stopped; don't continue
            if session_name and not stop_all and status == SessionStopStatus.SUCCESSFUL:
                break

    if warning_messages:
        for message in warning_messages:
            communication.warn(message)

    if not statues:
        return
    elif all(s == SessionStopStatus.NO_ACTIVE_SESSION for s in statues):
        raise errors.ParameterError("There are no running sessions.")
    elif session_name and not any(s == SessionStopStatus.SUCCESSFUL for s in statues):
        raise errors.ParameterError(f"Could not find '{session_name}' among the running sessions.")
    elif any(s == SessionStopStatus.FAILED for s in statues):
        raise errors.ParameterError("Cannot stop some sessions")
    elif not session_name and not any(s == SessionStopStatus.SUCCESSFUL for s in statues):
        raise errors.ParameterError("Session name is missing")


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def session_open(session_name: Optional[str], provider: Optional[str] = None, **kwargs):
    """Open interactive session in the browser.

    Args:
        session_name(Optional[str]): Name of the session to open.
        provider(Optional[str]): Name of the session provider to use.
    """
    providers = [_safe_get_provider(provider)] if provider else get_supported_session_providers()
    project_name = get_renku_project_name()

    for session_provider in providers:
        if session_provider.session_open(project_name, session_name, **kwargs):
            return

    if session_name:
        raise errors.ParameterError(f"Could not find '{session_name}' among the running sessions.")
    else:
        raise errors.ParameterError("Session name is missing")


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def ssh_setup(existing_key: Optional[Path] = None, force: bool = False):
    """Setup SSH keys for SSH connections to sessions.

    Args:
        existing_key(Path, optional): Existing private key file to use instead of generating new ones.
        force(bool): Whether to prompt before overwriting keys or not
    """

    if not shutil.which("ssh"):
        raise errors.SSHNotFoundError()

    system_config = SystemSSHConfig()

    include_string = f"Include {system_config.renku_ssh_root}/*.conf\n\n"

    if include_string not in system_config.ssh_config.read_text():
        with system_config.ssh_config.open(mode="r+") as f:
            content = f.read()
            f.seek(
                0, 0
            )  # NOTE: We need to add 'Include' before any 'Host' entry, otherwise it is included as part of a host
            f.write(include_string + content)

    if not existing_key and not force and system_config.is_configured:
        communication.confirm(f"Keys already configured for host {system_config.renku_host}. Overwrite?", abort=True)

    if existing_key:
        communication.info("Linking existing keys")
        existing_public_key = existing_key.parent / (existing_key.name + ".pub")

        if not existing_key.exists() or not existing_public_key.exists():
            raise errors.KeyNotFoundError(
                f"Couldn't find private key '{existing_key}' or public key '{existing_public_key}'."
            )

        if system_config.keyfile.exists():
            system_config.keyfile.unlink()
        if system_config.public_keyfile.exists():
            system_config.public_keyfile.unlink()

        os.symlink(existing_key, system_config.keyfile)
        os.symlink(existing_public_key, system_config.public_keyfile)
    else:
        communication.info("Generating keys")
        keys = generate_ssh_keys()
        system_config.keyfile.touch(mode=0o600)
        system_config.public_keyfile.touch(mode=0o644)
        with system_config.keyfile.open(
            "wt",
        ) as f:
            f.write(keys.private_key)

        with system_config.public_keyfile.open("wt") as f:
            f.write(keys.public_key)

    communication.info("Writing SSH config")
    with system_config.jumphost_file.open(mode="wt") as f:
        # NOTE: The * at the end of the jumphost name hides it from VSCode
        content = textwrap.dedent(
            f"""
            Host jumphost-{system_config.renku_host}*
                HostName {system_config.renku_host}
                Port 2022
                User jovyan
            """
        )
        f.write(content)

    communication.warn(
        "This command does not add any public SSH keys to your project. "
        "Keys have to be added manually or by using the 'renku session start' command with the '--ssh' flag."
    )


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def session_pause(session_name: Optional[str], provider: Optional[str] = None, **kwargs):
    """Pause an interactive session.

    Args:
        session_name(Optional[str]): Name of the session.
        provider(Optional[str]): Name of the session provider to use.
    """

    def pause(session_provider: IHibernatingSessionProvider) -> SessionStopStatus:
        try:
            return session_provider.session_pause(project_name=project_name, session_name=session_name)
        except errors.RenkulabSessionGetUrlError:
            if provider:
                raise
            return SessionStopStatus.FAILED

    project_name = get_renku_project_name()

    if provider:
        session_provider = _safe_get_provider(provider)
        if session_provider is None:
            raise errors.ParameterError(f"Provider '{provider}' not found")
        elif not isinstance(session_provider, IHibernatingSessionProvider):
            raise errors.ParameterError(f"Provider '{provider}' doesn't support pausing sessions")
        providers = [session_provider]
    else:
        providers = get_supported_hibernating_session_providers()

    session_message = f"session {session_name}" if session_name else "session"
    statues = []
    warning_messages = []
    with communication.busy(msg=f"Waiting for {session_message} to pause..."):
        for session_provider in sorted(providers, key=lambda p: p.priority):
            try:
                status = pause(session_provider)  # type: ignore
            except errors.RenkuException as e:
                warning_messages.append(f"Cannot pause sessions in provider '{session_provider.name}': {e}")
            else:
                statues.append(status)

            # NOTE: The given session name was stopped; don't continue
            if session_name and status == SessionStopStatus.SUCCESSFUL:
                break

    if warning_messages:
        for message in warning_messages:
            communication.warn(message)

    if not statues:
        return
    elif all(s == SessionStopStatus.NO_ACTIVE_SESSION for s in statues):
        raise errors.ParameterError("There are no running sessions.")
    elif session_name and not any(s == SessionStopStatus.SUCCESSFUL for s in statues):
        raise errors.ParameterError(f"Could not find '{session_name}' among the running sessions.")
    elif not session_name and not any(s == SessionStopStatus.SUCCESSFUL for s in statues):
        raise errors.ParameterError("Session name is missing")


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def session_resume(session_name: Optional[str], provider: Optional[str] = None, **kwargs):
    """Resume a paused session.

    Args:
        session_name(Optional[str]): Name of the session.
        provider(Optional[str]): Name of the session provider to use.
    """
    project_name = get_renku_project_name()

    if provider:
        session_provider = _safe_get_provider(provider)
        if session_provider is None:
            raise errors.ParameterError(f"Provider '{provider}' not found")
        elif not isinstance(session_provider, IHibernatingSessionProvider):
            raise errors.ParameterError(f"Provider '{provider}' doesn't support pausing/resuming sessions")
        providers = [session_provider]
    else:
        providers = get_supported_hibernating_session_providers()

    session_message = f"session {session_name}" if session_name else "session"
    with communication.busy(msg=f"Waiting for {session_message} to resume..."):
        for session_provider in providers:
            if session_provider.session_resume(project_name, session_name, **kwargs):  # type: ignore
                return

    if session_name:
        raise errors.ParameterError(f"Could not find '{session_name}' among the sessions.")
    else:
        raise errors.ParameterError("Session name is missing")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A factory to get various data providers."""

from typing import List, Type, Union
from urllib.parse import urlparse

from renku.core import errors
from renku.core.dataset.providers.api import (
    AddProviderInterface,
    ExportProviderInterface,
    ImportProviderInterface,
    ProviderApi,
    StorageProviderInterface,
)
from renku.core.plugin.dataset_provider import get_supported_dataset_providers
from renku.core.util import communication
from renku.core.util.doi import is_doi


# TODO: Fix return type of methods once python supports type intersections: python/typing/issues/213
class ProviderFactory:
    """Create a provider type from URI."""

    @staticmethod
    def get_providers() -> List[Type[ProviderApi]]:
        """Return a list of providers sorted based on their priorities (higher priority providers come first)."""
        providers = get_supported_dataset_providers()
        return sorted(providers, key=lambda p: p.priority.value)  # type: ignore

    @staticmethod
    def get_add_providers() -> List[Union[Type[ProviderApi], Type[AddProviderInterface]]]:
        """Get a list of dataset add providers."""
        return [p for p in ProviderFactory.get_providers() if issubclass(p, AddProviderInterface)]

    @staticmethod
    def get_export_providers() -> List[Union[Type[ProviderApi], Type[ExportProviderInterface]]]:
        """Get a list of dataset exporter providers."""
        return [p for p in ProviderFactory.get_providers() if issubclass(p, ExportProviderInterface)]

    @staticmethod
    def get_import_providers() -> List[Union[Type[ProviderApi], Type[ImportProviderInterface]]]:
        """Get a list of dataset importer providers."""
        return [p for p in ProviderFactory.get_providers() if issubclass(p, ImportProviderInterface)]

    @staticmethod
    def get_storage_providers() -> List[Union[Type[ProviderApi], Type[StorageProviderInterface]]]:
        """Get a list of backend storage providers."""
        return [p for p in ProviderFactory.get_providers() if issubclass(p, StorageProviderInterface)]

    @staticmethod
    def get_add_provider(uri):
        """Get an add provider based on uri."""
        for provider in ProviderFactory.get_add_providers():
            try:
                supports = provider.supports(uri)  # type: ignore[union-attr]
            except BaseException as e:
                communication.warn(f"Couldn't test add provider {provider.__class__.__name__}: {e}")
            else:
                if supports:
                    return provider(uri=uri)  # type: ignore[call-arg]

        raise errors.DatasetProviderNotFound(uri=uri)

    @staticmethod
    def get_export_provider(provider_name):
        """Get the export provider with a given name."""
        provider_name = provider_name.lower()
        try:
            provider = next(
                p for p in ProviderFactory.get_export_providers() if p.name.lower() == provider_name  # type: ignore
            )
            return provider(uri=None)  # type: ignore
        except StopIteration:
            raise errors.DatasetProviderNotFound(name=provider_name)

    @staticmethod
    def get_import_provider(uri):
        """Get an import provider based on uri."""
        is_doi_ = is_doi(uri)
        if not is_doi_:
            url = urlparse(uri)
            if bool(url.scheme and url.netloc and url.params == "") is False:
                raise errors.DatasetProviderNotFound(message=f"Cannot parse URL '{uri}'")

        warning = ""
        import_providers = ProviderFactory.get_import_providers()

        for provider in import_providers:
            try:
                supports = provider.supports(uri)  # type: ignore[union-attr]
            except BaseException as e:
                warning += f"Couldn't test provider {provider}: {e}\n"
            else:
                if supports:
                    return provider(uri=uri, is_doi=is_doi_)  # type: ignore[call-arg]

        url = uri.split("/")[1].split(".")[0] if is_doi_ else uri  # NOTE: Get DOI provider name if uri is a DOI
        supported_providers = ", ".join(p.name for p in import_providers)  # type: ignore
        message = warning + f"Provider not found: {url}\nHint: Supported providers are: {supported_providers}"
        raise errors.DatasetProviderNotFound(message=message)

    @staticmethod
    def get_storage_provider(uri):
        """Get a backend storage provider based on uri."""
        for provider in ProviderFactory.get_storage_providers():
            try:
                supports = provider.supports_storage(uri)  # type: ignore[union-attr]
            except BaseException as e:
                communication.warn(f"Couldn't test provider {provider}: {e}")
            else:
                if supports:
                    return provider(uri=uri)  # type: ignore[call-arg]

        raise errors.DatasetProviderNotFound(uri=uri)

    get_create_provider = get_storage_provider
    get_mount_provider = get_storage_provider
    get_pull_provider = get_storage_provider



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service datasets serializers."""
from marshmallow import Schema, ValidationError, fields, post_load

from renku.domain_model.dataset import DatasetCreatorsJson as DatasetCreators
from renku.domain_model.dataset import DatasetDetailsJson as DatasetDetails
from renku.domain_model.dataset import ImageObjectJson as ImageObject
from renku.domain_model.dataset import ImageObjectRequestJson
from renku.ui.service.serializers.common import (
    AsyncSchema,
    GitCommitSHA,
    GitUrlResponseMixin,
    JobDetailsResponse,
    MigrateSchema,
    RemoteRepositorySchema,
    RenkuSyncSchema,
)
from renku.ui.service.serializers.rpc import JsonRPCResponse


class DatasetSlugSchema(Schema):
    """Schema for dataset slug."""

    slug = fields.String(metadata={"description": "Mandatory dataset slug."}, required=True)


class DatasetDetailsRequest(DatasetDetails):
    """Request schema with dataset image information."""

    images = fields.List(fields.Nested(ImageObjectRequestJson))

    custom_metadata: fields.Field = fields.Dict()


class DatasetCreateRequest(AsyncSchema, DatasetDetailsRequest, RemoteRepositorySchema, MigrateSchema):
    """Request schema for a dataset create view."""

    # NOTE: Override field in DatasetDetails
    data_directory = fields.String(  # type: ignore
        load_default=None,
        metadata={"description": "Base dataset data directory. '<project.data_directory>/<dataset.slug>' by default"},
    )


class DatasetCreateResponse(DatasetSlugSchema, RenkuSyncSchema, GitUrlResponseMixin):
    """Response schema for a dataset create view."""


class DatasetCreateResponseRPC(JsonRPCResponse):
    """RPC response schema for dataset create view."""

    result = fields.Nested(DatasetCreateResponse)


class DatasetRemoveRequest(AsyncSchema, DatasetSlugSchema, RemoteRepositorySchema, MigrateSchema):
    """Request schema for a dataset remove."""


class DatasetRemoveResponse(DatasetSlugSchema, RenkuSyncSchema, GitUrlResponseMixin):
    """Response schema for a dataset create view."""


class DatasetRemoveResponseRPC(JsonRPCResponse):
    """RPC response schema for dataset create view."""

    result = fields.Nested(DatasetRemoveResponse)


class DatasetAddFile(Schema):
    """Schema for a dataset add file view."""

    file_id = fields.String()
    file_path = fields.String()
    file_url = fields.String()
    job_id = fields.String()


class DatasetAddRequest(AsyncSchema, DatasetSlugSchema, RemoteRepositorySchema, MigrateSchema):
    """Request schema for a dataset add file view."""

    files = fields.List(fields.Nested(DatasetAddFile), required=True)

    create_dataset = fields.Boolean(load_default=False)
    force = fields.Boolean(load_default=False)

    client_extras = fields.String()

    @post_load()
    def check_files(self, data, **kwargs):
        """Check serialized file list."""
        for _file in data["files"]:
            if "file_id" in _file and "file_path" in _file:
                raise ValidationError("invalid reference found: use either `file_id` or `file_path`")

        return data


class DatasetAddResponse(DatasetSlugSchema, RenkuSyncSchema, GitUrlResponseMixin):
    """Response schema for a dataset add file view."""

    project_id = fields.String(required=True)
    files = fields.List(fields.Nested(DatasetAddFile), required=True)


class DatasetAddResponseRPC(JsonRPCResponse):
    """RPC schema for a dataset add."""

    result = fields.Nested(DatasetAddResponse)


class DatasetListRequest(RemoteRepositorySchema, GitCommitSHA):
    """Request schema for dataset list view."""


class DatasetDetailsResponse(DatasetDetails):
    """Request schema with dataset image information."""

    images = fields.List(fields.Nested(ImageObject))


class DatasetListResponse(GitUrlResponseMixin):
    """Response schema for dataset list view."""

    datasets = fields.List(fields.Nested(DatasetDetailsResponse), required=True)


class DatasetListResponseRPC(JsonRPCResponse):
    """RPC response schema for dataset list view."""

    result = fields.Nested(DatasetListResponse)


class DatasetFilesListRequest(DatasetSlugSchema, RemoteRepositorySchema, GitCommitSHA):
    """Request schema for dataset files list view."""


class DatasetFileDetails(Schema):
    """Serialize dataset files to a response object."""

    name = fields.String(metadata={"description": "Mandatory dataset file name."}, required=True)
    path = fields.String()
    created = fields.DateTime()
    added = fields.DateTime()


class DatasetFilesListResponse(DatasetSlugSchema, GitUrlResponseMixin):
    """Response schema for dataset files list view."""

    files = fields.List(fields.Nested(DatasetFileDetails), required=True)


class DatasetFilesListResponseRPC(JsonRPCResponse):
    """RPC schema for dataset files list view."""

    result = fields.Nested(DatasetFilesListResponse)


class DatasetImportRequest(AsyncSchema, RemoteRepositorySchema, MigrateSchema):
    """Dataset import request."""

    dataset_uri = fields.String(required=True)
    slug = fields.String(metadata={"description": "Optional dataset slug."})
    extract = fields.Boolean()
    tag = fields.String(metadata={"description": "Dataset version to import."})
    data_directory = fields.String(
        load_default=None,
        metadata={"description": "Base dataset data directory. '<project.data_directory>/<dataset.slug>' by default"},
    )


class DatasetImportResponseRPC(JsonRPCResponse):
    """RPC schema for a dataset import."""

    result = fields.Nested(JobDetailsResponse)


class DatasetEditRequest(
    AsyncSchema,
    DatasetDetailsRequest,
    DatasetSlugSchema,
    RemoteRepositorySchema,
    MigrateSchema,
):
    """Dataset edit metadata request."""

    name = fields.String(metadata={"description": "New name of the dataset"})
    description = fields.String(metadata={"description": "New description of the dataset"})
    creators = fields.List(fields.Nested(DatasetCreators), metadata={"description": "New creators of the dataset"})
    keywords = fields.List(fields.String(), allow_none=True, metadata={"description": "New keywords for the dataset"})
    images = fields.List(
        fields.Nested(ImageObjectRequestJson), allow_none=True, metadata={"description": "New dataset images"}
    )
    custom_metadata = fields.List(
        fields.Dict(), metadata={"description": "New custom metadata for the dataset"}, allow_none=True
    )
    custom_metadata_source = fields.String(
        allow_none=True,
        metadata={"description": "Source for the custom metadata for the dataset"},
    )


class DatasetEditResponse(RenkuSyncSchema, GitUrlResponseMixin):
    """Dataset edit metadata response."""

    edited = fields.Dict(required=True)
    warnings = fields.List(fields.String())


class DatasetEditResponseRPC(JsonRPCResponse):
    """RPC schema for a dataset import."""

    result = fields.Nested(DatasetEditResponse)


class DatasetUnlinkRequest(AsyncSchema, DatasetSlugSchema, RemoteRepositorySchema, MigrateSchema):
    """Dataset unlink file request."""

    include_filters = fields.List(fields.String())
    exclude_filters = fields.List(fields.String())

    @post_load()
    def check_filters(self, data, **kwargs):
        """Check filters."""
        include_filter = data.get("include_filters")
        exclude_filter = data.get("exclude_filters")

        if not include_filter and not exclude_filter:
            raise ValidationError("one of the filters must be specified")

        return data


class DatasetUnlinkResponse(RenkuSyncSchema, GitUrlResponseMixin):
    """Dataset unlink files response."""

    unlinked = fields.List(fields.String())


class DatasetUnlinkResponseRPC(JsonRPCResponse):
    """Dataset unlink files RPC response."""

    result = fields.Nested(DatasetUnlinkResponse)
